#ifndef __AOBJECT_H
#define __AOBJECT_H

enum AstObjectFlag_e
{
  ofVisible    = 0x0001,
  ofValid      = 0x0002,
  ofSolid      = 0x0100
};

const int MAX_OBJECT_NAME = 40;

class AstObject
{
  friend class AstWorld;

  protected:
    char name[MAX_OBJECT_NAME];

    AstObjectDefinition *curdef;
    AstObjectDefinition *fardef, *neardef;
    AstFixedpoint_t      neardef_Z, fardef_Z;

    AstObject       *next_in_world;

    AstVector       *scratch_vertex_array;
    AstScreenVertex *scratch_screen_array;
    bool            *scratch_cull_array;
    AstFixedpoint_t *scratch_intensity_array;

    AstMatrix4x3 world_to_object_matrix;
    AstMatrix4x3 local_matrix;
    AstVector    location_CCS;

    union {
       struct {
          unsigned visible    : 1,
                   valid      : 1,
                   reserved   : 6,
                   solid      : 1;
       } aflags_s;
       unsigned oflags;
    };

    AstFixedpoint_t  bounding_sphere;
    long             bounding_sphere_squared;

    virtual void   CalcExtents( void );

    bool   CheckProjectedBoundingSphere( void );
    long   DistanceSquaredTo( const AstObject &o ) const;
    bool   SelectRepresentation( void );

    void   CullAndShade( void );
    void   DrawTriangles( void );
    void   Shade( void );
    void   ProjectVertices( void );

    void   AllocateScratchArrays( void );
    void   FreeScratchArrays( void );
    void   TransformVerticesToCCS( void );

    AstVector GetLocationCCS( void ) const { return location_CCS; }

    AstFixedpoint_t   GetBoundingSphere( void ) const { return bounding_sphere; }
    long              GetBoundingSphereSquared( void ) const { return bounding_sphere_squared; }

  public:
    AstObject( const char *name = 0 );
    virtual ~AstObject();

    const char *GetName( void ) const { return name; }
    void        SetName( const char *n ) { strcpy( name, n ); }

    const AstMatrix4x3 *CalculateWorldToObjectMatrix();

    AstVector GetPositionCCS( void ) const { return location_CCS; }
    AstVector GetPositionWCS( void ) const { return local_matrix.GetTranslation(); }

    void      MoveAlongX( AstFixedpoint_t amount );
    void      MoveAlongY( AstFixedpoint_t amount );
    void      MoveAlongZ( AstFixedpoint_t amount );
    void      MoveTo( const AstVector &newposition )
    {
       local_matrix.SetTranslation( newposition );
    }
    void      MoveTo( AstFixedpoint_t x, AstFixedpoint_t y, AstFixedpoint_t z )
    {
        MoveTo( AstVector( x, y, z ) );
    }

    void      RotateAroundX( unsigned short degrees );
    void      RotateAroundY( unsigned short degrees );
    void      RotateAroundZ( unsigned short degrees );

    virtual bool DetermineVisibility( void );
    virtual void DoSomething( void ) {}
    virtual void Draw( void );

    virtual void YouHit( AstObject * ) {}

    bool CheckFlag( AstObjectFlag_e bit ) const
    {
       return ( oflags & bit ) ? true : false;
    }
    void SetFlag( AstObjectFlag_e flag, bool value )
    {
       if ( value ) oflags |= flag;
       else oflags &= ~flag;
    }

    void SetFarDefinition( AstObjectDefinition *d, AstFixedpoint_t Z )
    {
       fardef = d;
       fardef_Z = Z;
    }
    void SetNearDefinition( AstObjectDefinition *d, AstFixedpoint_t Z )
    {
       neardef = d;
       neardef_Z = Z;
       AllocateScratchArrays();
       CalcExtents();
    }

    bool Intersects( const AstObject *object ) const;
};

#endif
