#include "ast3d.h"

/*
** static bool SeekToSection( FILE *fp, const char *section_name )
**
** Helper function used for getting around DEF files.
*/
static bool SeekToSection( FILE *fp, const char *section_name )
{
char buffer[80];

   rewind( fp );
   while ( !feof( fp ) ) {
      if ( fgets( buffer, 80, fp ) ) {
         if ( strchr( buffer, '\n' ) )
            *strchr( buffer, '\n' ) = 0;
         if ( strcmp( buffer, section_name ) == 0 )
            return true;
      }
      else {
         return false;
      }
   }
   return false;
}

/*
** static int CountEntries( FILE *fp, const char *section_name )
**
** Helper function that counts the # of valid entries in a section.
** This is not the most robust routine, but it gets the job done.
*/
static int CountEntries( FILE *fp, const char *section_name )
{
char buffer[80];
int  sum = 0;

   /*
   ** Seek to the section name
   */
   if ( !SeekToSection( fp, section_name ) )
      AstFatal( "Could not find section in DEF file!" );

   /*
   ** Count each line that does not begin with '[', '\n', or is
   ** NULL as a valid entry, even thought it may not be.
   */
   while ( !feof( fp ) ) {
      if ( fgets( buffer, 80, fp ) ) {
         if ( buffer[0] == '[' || buffer[0] == '\n' || buffer[0] == 0 )
            break;
         else
            sum++;
      }
      else {
         break;
      }
   }
   return sum;
}

/*
** AstObjectDefinition::AstObjectDefinition( const char *filename, AstFixedpoint_t sd )
**
** This is the constructor for the AstObjectDefinition class.  The filename
** passed is the name of the DEF file from which the AstObjectDefinition
** will read its characteristics.  The "sd" value is a scaling value used
** for scaling some files.  This is necessary with some files, such as
** converted PLG files that operate in a completely different magnitude
** space than AST.
*/
AstObjectDefinition::AstObjectDefinition( const char *filename, AstFixedpoint_t sd )
{
   triangle_array = 0;
   vertex_array   = 0;
   num_triangles  = 0;
   num_vertices   = 0;

   valid = LoadFrom( filename, sd );
}

/*
** AstObjectDefinition::~AstObjectDefinition()
**
** This is the destructor for the AstObjectDefinition class.  It
** frees up memory associated with the AstObjectDefinition.
*/
AstObjectDefinition::~AstObjectDefinition()
{
   delete []vertex_array;
   vertex_array = 0;
   delete []triangle_array;
   triangle_array = 0;
   delete []texture_vertex_array;
   texture_vertex_array = 0;
}

/*
** void AstObjectDefinition::AssignTextureMap( AstTextureMap *map )
**
** Assigns "map" as the texture map for all triangles.  Obviously this
** limits the flexibility of texture mapping in AST3D, since an object
** must be completely texture mapped with the same texture map.  This
** limitation is fairly trivial to compensate for by reading in texture
** map names as part of the triangle definitions.
*/
void AstObjectDefinition::AssignTextureMap( AstTextureMap *map )
{
   for ( int i = 0; i < num_triangles; i++ )
      triangle_array[i].map = map;
}

/*
** bool AstObjectDefinition::LoadFrom( const char *fname, AstFixedpoint_t scale_divisor )
**
** This routine does the brunt of the object definition file reading
** and definition stuff.
*/
bool AstObjectDefinition::LoadFrom( const char *filename, AstFixedpoint_t scale_divisor )
{
FILE *fp;
int   i;

   /*
   ** Open the file
   */
   if ( ( fp = fopen( filename, "rt" ) ) == 0 )
      return false;

   /*
   ** See if the ordering is clockwise or counter clockwise
   */
   if ( SeekToSection( fp, "[ordering=cw]" ) )
      ccw = false;
   else
      ccw = true;

   /*
   ** Load texture vertices
   */
   num_texture_vertices = CountEntries( fp, "[texture vertices]") ;
   LoadTextureVertices( fp );

   num_vertices = CountEntries( fp, "[vertices]" );
   LoadVertices( fp );

   /*
   ** Apply a scaling factor if necessary
   */
   for ( i = 0; i < num_vertices; i++ ) {
      vertex_array[i].X = AstFixDivFix( vertex_array[i].X, scale_divisor );
      vertex_array[i].Y = AstFixDivFix( vertex_array[i].Y, scale_divisor );
      vertex_array[i].Z = AstFixDivFix( vertex_array[i].Z, scale_divisor );
   }

   /*
   ** Load the triangles
   */
   num_triangles = CountEntries( fp, "[triangles]" );
   LoadTriangles( fp );

   /*
   ** Then precompute the plane constants
   */
   for ( i = 0; i < num_triangles; i++ )
      triangle_array[i].ComputePlaneConstant( *this );

   fclose( fp );
   return true;
}

/*
** void AstObjectDefinition::LoadTextureVertices( FILE *fp )
**
** This function loads the texture vertices from using the
** given FILE pointer.
*/
void AstObjectDefinition::LoadTextureVertices( FILE *fp )
{
char buffer[80];
bool done = false;
int  cur_tv = 0;

   /*
   ** Seek to the correct section
   */
   if ( !SeekToSection( fp, "[texture vertices]" ) )
      AstFatal( "Could not find [texture vertices] marker in DEF file!" );

   /*
   ** Allocate memory
   */
   if ( ( texture_vertex_array = new AstScreenVertex[num_texture_vertices] ) == 0 )
      AstFatal( "texture_vertex_array could not be allocated in AstObjectDefinition::LoadTriangles()!" );

   /*
   ** Do the deed
   */
   while ( !done ) {
      if ( feof( fp ) ) {
         done = true;
      }
      else {
         if ( fgets( buffer, 80, fp ) ) {
            if ( buffer[0] == '[' ) {
               done = true;
            }
            else if ( buffer[0] != 0 && buffer[0] != '\n' ) {
               AstScreenVertex tv;

               if ( strchr( buffer, '\n' ) ) *strchr( buffer, '\n' ) = 0;

               sscanf( buffer, "%d,%d", &tv.X, &tv.Y );

               texture_vertex_array[cur_tv] = tv;

               cur_tv++;
            }
         }
         else {
            done = true;
         }
      }
   }
}

/*
** void AstObjectDefinition::LoadTriangles( FILE *fp )
**
** Loads triangles from the given FILE pointer.
*/
void AstObjectDefinition::LoadTriangles( FILE *fp )
{
char buffer[80];
bool done = false;
int  cur_tri = 0;

   /*
   ** As usual, seek to the correct section first
   */
   if ( !SeekToSection( fp, "[triangles]" ) )
      AstFatal( "Could not find [triangles] marker in DEF file!" );

   /*
   ** Allocate memory
   */
   if ( ( triangle_array = new AstTriangle[num_triangles] ) == 0 )
      AstFatal( "triangle_array could not be allocated in AstObjectDefinition::LoadTriangles()!" );

   /*
   ** Do it
   */
   while ( !done ) {
      if ( feof( fp ) ) {
         done = true;
      }
      else {
         if ( fgets( buffer, 80, fp ) ) {
            if ( buffer[0] == '[' ) {
               done = true;
            }
            else if ( buffer[0] != 0 && buffer[0] != '\n' ) {
               int    a, b, c, tex_index[3], col;
               TtType type;

               if ( strchr( buffer, '\n' ) ) *strchr( buffer, '\n' ) = 0;

               sscanf( buffer, "%d,%d,%d - %d,%d,%d:%d,%d", &a, &b, &c, &tex_index[0], &tex_index[1], &tex_index[2], &type, &col );

               /*
               ** Correctly order depending on orientation
               */
               if ( ccw ) {
                  triangle_array[cur_tri].vertices[0] = a;
                  triangle_array[cur_tri].vertices[1] = b;
                  triangle_array[cur_tri].vertices[2] = c;
                  triangle_array[cur_tri].texture_vertices[0] = texture_vertex_array[tex_index[0]];
                  triangle_array[cur_tri].texture_vertices[1] = texture_vertex_array[tex_index[1]];
                  triangle_array[cur_tri].texture_vertices[2] = texture_vertex_array[tex_index[2]];
               }
               else {
                  triangle_array[cur_tri].vertices[0] = c;
                  triangle_array[cur_tri].vertices[1] = b;
                  triangle_array[cur_tri].vertices[2] = a;
                  triangle_array[cur_tri].texture_vertices[0] = texture_vertex_array[tex_index[2]];
                  triangle_array[cur_tri].texture_vertices[1] = texture_vertex_array[tex_index[1]];
                  triangle_array[cur_tri].texture_vertices[2] = texture_vertex_array[tex_index[0]];
               }
               triangle_array[cur_tri].color       = col;
               triangle_array[cur_tri].type        = type;

               cur_tri++;
            }
         }
         else {
            done = true;
         }
      }
   }
}

/*
** void AstObjectDefinition::LoadVertices( FILE *fp )
**
** Load vertices for the object definition using the given
** FILE pointer.
*/
void AstObjectDefinition::LoadVertices( FILE *fp )
{
char buffer[80];
bool done = false;
int  cur_vtx = 0;

   /*
   ** Seek to the right section
   */
   if ( !SeekToSection( fp, "[vertices]" ) )
      AstFatal( "Could not find [vertices] marker in DEF file!" );

   /*
   ** Allocate memory
   */
   if ( ( vertex_array = new AstVector[num_vertices] ) == 0 )
      AstFatal( "vertex_array could not be allocated in AstObjectDefinition::LoadVertices()!" );

   /*
   ** Do it
   */
   while ( !done ) {
      if ( feof( fp ) ) {
         done = true;
      }
      else {
         if ( fgets( buffer, 80, fp ) ) {
            if ( buffer[0] == '[' ) {
               done = true;
            }
            else if ( buffer[0] != 0 && buffer[0] != '\n' ) {
               float x, y, z;
               sscanf( buffer, "%f,%f,%f", &x, &y, &z );
               vertex_array[cur_vtx].X = AstConvertToFix( x );
               vertex_array[cur_vtx].Y = AstConvertToFix( y );
               vertex_array[cur_vtx].Z = AstConvertToFix( z );

               cur_vtx++;
            }
         }
         else {
            done = true;
         }
      }
   }
}
